#include <vector>
#include <algorithm>
#include <random>
#include <functional>
#include <iostream>

#include <omp.h>
#include <memory.h>

#pragma region OMP Radix
// https://haichuanwang.wordpress.com/2014/05/26/a-faster-openmp-radix-sort-implementation/
#define BASE_BITS 8
#define BASE (1 << BASE_BITS)
#define MASK (BASE-1)
#define DIGITS(v, shift) (((v) >> shift) & MASK)

void omp_lsd_radix_sort(size_t n, unsigned int *data) {
	unsigned int* buffer = new unsigned int[n];
	int total_digits = sizeof(unsigned int) * 8;

	//Each thread use local_bucket to move data
	int i;
	for (int shift = 0; shift < total_digits; shift += BASE_BITS) {
		size_t bucket[BASE] = { 0 };

		size_t local_bucket[BASE] = { 0 }; // size needed in each bucket/thread
						   //1st pass, scan whole and check the count
#pragma omp parallel firstprivate(local_bucket)
		{
#pragma omp for schedule(static) nowait
			for (i = 0; i < n; i++) {
				local_bucket[DIGITS(data[i], shift)]++;
			}
#pragma omp critical
			for (i = 0; i < BASE; i++) {
				bucket[i] += local_bucket[i];
			}
#pragma omp barrier
#pragma omp single
			for (i = 1; i < BASE; i++) {
				bucket[i] += bucket[i - 1];
			}
			int nthreads = omp_get_num_threads();
			int tid = omp_get_thread_num();
			for (int cur_t = nthreads - 1; cur_t >= 0; cur_t--) {
				if (cur_t == tid) {
					for (i = 0; i < BASE; i++) {
						bucket[i] -= local_bucket[i];
						local_bucket[i] = bucket[i];
					}
				}
				else { //just do barrier
#pragma omp barrier
				}

			}
#pragma omp for schedule(static)
			for (i = 0; i < n; i++) { //note here the end condition
				buffer[local_bucket[DIGITS(data[i], shift)]++] = data[i];
			}
		}
		//now move data
		unsigned int* tmp = data;
		data = buffer;
		buffer = tmp;
	}
	delete[] buffer;
}
#pragma endregion

#pragma region bubble sort
// bubble sort
void bubbleSort(size_t n, unsigned int* data)
{
	bool swapped;
	do {
		swapped = false;
		for (int k = 0; k + 1 < n; ++k)
		{
			if (data[k] > data[k + 1])
			{
				std::swap(data[k], data[k + 1]);
				swapped = true;
			}
		}
	} while (swapped);
}
#pragma endregion

#pragma region odd-even sort
// - stop condition: omp reduction clause
void oddEvenSort(size_t n, unsigned int* data)
{
	int oddeven = 0;
	bool swapped;
	do {
		swapped = false;
		for (int k = oddeven; k < n-1; k += 2) {
			if (data[k] > data[k + 1]) {
				std::swap(data[k], data[k + 1]);
				swapped = true;
			}
		}
		oddeven = 1 - oddeven;
	} while (swapped);

}
#pragma endregion

// Hybrid parallel sort:
// - local block sort
// - merge sort (binary tree - lower/upper half merge)
#pragma region merge
void merge(size_t n, unsigned int* src1, unsigned int* src2, unsigned int* dest)
{
	size_t isrc1 = 0, isrc2 = 0, idest = 0;
	while (isrc1 < n && isrc2 < n)
	{
		if (src1[isrc1] < src2[isrc2])
		{
			dest[idest++] = src1[isrc1++];
		}
		else
		{
			dest[idest++] = src2[isrc2++];
		}
	}
	while (isrc1 < n)
	{
		dest[idest++] = src1[isrc1++];
	}
	while (isrc2 < n)
	{
		dest[idest++] = src2[isrc2++];
	}
}

void mergeLowerHalf(size_t n, unsigned int* src1, unsigned int* src2, unsigned int* dest)
{
	size_t isrc1 = 0, isrc2 = 0, idest = 0;
	while (idest < n)
	{
		if (src1[isrc1] < src2[isrc2])
		{
			dest[idest++] = src1[isrc1++];
		}
		else
		{
			dest[idest++] = src2[isrc2++];
		}
	}
}

void mergeUpperHalf(size_t n, unsigned int* src1, unsigned int* src2, unsigned int* dest)
{
	size_t isrc1 = n - 1, isrc2 = n - 1, idest = n - 1;
	while (idest >= 0)
	{
		if (src1[isrc1] > src2[isrc2])
		{
			dest[idest--] = src1[isrc1--];
		}
		else
		{
			dest[idest--] = src2[isrc2--];
		}
	}
}
#pragma endregion

// Bucket sort:
// - partition data into _n_ buckets
// - distribute the _i_th bucket to _i_th node
// - local sort
// - collect the sorted buckets from the nodes

#pragma region quicksort
void partition(size_t lower, size_t upper, size_t *pivot, unsigned int* data)
{
	unsigned int x = data[lower];
	size_t up = lower + 1;
	size_t down = upper;
	while (up < down)
	{
		while ((up < down) && (data[up] <= x)) up++;
		while ((up < down) && (data[down] > x)) down--;
		if (up == down) break;
		std::swap(data[up], data[down]);
	}
	if (data[up] > x) up--;
	data[lower] = data[up]; data[up] = x;
	*pivot = up;
}

void quicksort(size_t begin, size_t end, unsigned int* data)
{
	if (begin < end)
	{
		size_t pivot;
		partition(begin, end, &pivot, data);
		quicksort(begin, pivot-1, data);
		quicksort(pivot+1, end, data);
	}
}
#pragma endregion

int main()
{
	std::default_random_engine generator;
	std::uniform_int_distribution<unsigned int> distribution(1, 1000);
	auto randgen = std::bind(distribution, generator);
	size_t elementNumber = 100;

	std::vector<unsigned int> randomData;
	for (size_t element = 0; element < elementNumber; ++element)
	{
		unsigned int i = randgen();
		randomData.push_back(randgen());
	}

	bubbleSort(randomData.size(), randomData.data());
	//oddEvenSort(randomData.size(), randomData.data());
	//quicksort(0, randomData.size(), randomData.data());
	//std::sort(randomData.begin(), randomData.end());
	//omp_lsd_radix_sort(randomData.size(), randomData.data());

	if (elementNumber <= 100)
	{
		for (int i = 0; i < elementNumber; ++i)
		{
			std::cout << randomData[i] << ",";
		}
		std::cout << std::endl;
	}
	return 0;
}

